/*
 * File         : ExpressionFindingVisitor.java
 * Author       : Guansong Zhang
 * Created      : Mon Oct 19 02:26:24 1998
 * Revision     : $Revision: 1.2 $
 * Updated      : $Date: 1999/10/25 22:07:57 $
 */

package hpjava.compiler.abstractexp; 

import hpjava.compiler.syntaxtree.*; 
import hpjava.compiler.visitor.*; 

public class ExpressionFindingVisitor extends DepthFirstVisitor {
  // override visitor to find expression faster
   //
   // User-generated visitor methods below
   //

   //
   // f0 -> [ PackageDeclaration() ]
   // f1 -> ( ImportDeclaration() )*
   // f2 -> ( TypeDeclaration() )*
   // f3 -> <EOF>
   //
   public void visit(CompilationUnit n) {
      n.f2.accept(this);
      //      out.println("");
   }

   //
   // f0 -> ClassDeclaration()
   //       | InterfaceDeclaration()
   //       | ";"
   //
   public void visit(TypeDeclaration n) {
      if(n.f0.which != 2){
        n.f0.accept(this);
      }
   }

   //
   // f0 -> ( "abstract" | "final" | "public" )*
   // f1 -> UnmodifiedClassDeclaration()
   //
   public void visit(ClassDeclaration n) {
      n.f1.accept(this);
   }

   //
   // f0 -> "class"
   // f1 -> <IDENTIFIER>
   // f2 -> [ "extends" Name() ]
   // f3 -> [ "implements" NameList() ]
   // f4 -> ClassBody()
   //
   public void visit(UnmodifiedClassDeclaration n) {
      n.f4.accept(this);
   }

   //
   // f0 -> "{"
   // f1 -> ( ClassBodyDeclaration() )*
   // f2 -> "}"
   //
   public void visit(ClassBody n) {
      n.f1.accept(this);
   }

   //
   // f0 -> ( "static" | "abstract" | "final" | "public" | "protected" | "private" )*
   // f1 -> UnmodifiedClassDeclaration()
   //
   public void visit(NestedClassDeclaration n) {
      n.f1.accept(this);
   }

   //
   // f0 -> Initializer()
   //       | NestedClassDeclaration()
   //       | NestedInterfaceDeclaration()
   //       | ConstructorDeclaration()
   //       | MethodDeclaration()
   //       | FieldDeclaration()
   //       | ";"
   //
   public void visit(ClassBodyDeclaration n) {
      n.f0.accept(this);
   }

   //
   // f0 -> ( "abstract" | "public" )*
   // f1 -> UnmodifiedInterfaceDeclaration()
   //
   public void visit(InterfaceDeclaration n) {
      n.f1.accept(this);
   }

   //
   // f0 -> ( "static" | "abstract" | "final" | "public" | "protected" | "private" )*
   // f1 -> UnmodifiedInterfaceDeclaration()
   //
   public void visit(NestedInterfaceDeclaration n) {
      n.f1.accept(this);
   }

   //
   // f0 -> "interface"
   // f1 -> <IDENTIFIER>
   // f2 -> [ "extends" NameList() ]
   // f3 -> "{"
   // f4 -> ( InterfaceMemberDeclaration() )*
   // f5 -> "}"
   //
   public void visit(UnmodifiedInterfaceDeclaration n) {
      n.f4.accept(this);
   }

   //
   // f0 -> NestedClassDeclaration()
   //       | NestedInterfaceDeclaration()
   //       | MethodDeclaration()
   //       | FieldDeclaration()
   //       | ";"
   //
   public void visit(InterfaceMemberDeclaration n) {
      if(n.f0.which != 4){
        n.f0.accept(this);
      }
   }

   //
   // f0 -> ( "public" | "protected" | "private" | "static" | "final" | "transient" | "volatile" )*
   // f1 -> Type()
   // f2 -> VariableDeclarator()
   // f3 -> ( "," VariableDeclarator() )*
   // f4 -> ";"
   //
   public void visit(FieldDeclaration n) {
      n.f2.accept(this);

      if(n.f3.present()){
        for(int i=0; i < n.f3.size(); i++){
          NodeSequence seq = (NodeSequence)n.f3.elementAt(i);
          seq.elementAt(1).accept(this);
        }
      }
   }

   //
   // f0 -> VariableDeclaratorId()
   // f1 -> [ "=" VariableInitializer() ]
   //
   public void visit(VariableDeclarator n) {
      if( n.f1.present() ){
        (((NodeSequence)n.f1.node).elementAt(1)).accept(this);
      }
   }

   //
   // f0 -> ArrayInitializer()
   //       | Expression()
   //
   public void visit(VariableInitializer n) {
      n.f0.accept(this);
   }

   //
   // f0 -> "{"
   // f1 -> [ VariableInitializer() ( "," VariableInitializer() )* ]
   // f2 -> [ "," ]
   // f3 -> "}"
   //
   public void visit(ArrayInitializer n) {
      if(n.f1.present()){
        NodeSequence seq1 = (NodeSequence)n.f1.node;
        seq1.elementAt(0).accept(this);
 
        if(((NodeListOptional)seq1.elementAt(1)).present()){
          NodeListOptional nlo = (NodeListOptional)seq1.elementAt(1);
          for(int j=0; j<nlo.size(); j++){
            NodeSequence seq2 = (NodeSequence)nlo.elementAt(j);
            seq2.elementAt(1).accept(this);
          }
        }
      }

   }

   //
   // f0 -> ( "public" | "protected" | "private" | "static" | "abstract" | "final" | "native" | "synchronized" )*
   // f1 -> ResultType()
   // f2 -> MethodDeclarator()
   // f3 -> [ "throws" NameList() ]
   // f4 -> ( Block() | ";" )
   //
   public void visit(MethodDeclaration n) {
      if(n.f4.which == 0){
        n.f4.accept(this);
      }
   }

   //
   // f0 -> [ "public" | "protected" | "private" ]
   // f1 -> <IDENTIFIER>
   // f2 -> FormalParameters()
   // f3 -> [ "throws" NameList() ]
   // f4 -> "{"
   // f5 -> [ ExplicitConstructorInvocation() ]
   // f6 -> ( BlockStatement() )*
   // f7 -> "}"
   //
   public void visit(ConstructorDeclaration n) {
      n.f5.accept(this);
      n.f6.accept(this);
   }

   //
   // f0 -> "this" Arguments() ";"
   //       | [ PrimaryExpression() "." ] "super" Arguments() ";"
   //
   public void visit(ExplicitConstructorInvocation n) {
      if(n.f0.which == 0){
        NodeSequence seq = (NodeSequence)n.f0.choice;
        seq.elementAt(1).accept(this);
      }
      else{
        NodeSequence seq = (NodeSequence)n.f0.choice;
        seq.elementAt(0).accept(this);
        seq.elementAt(2).accept(this);
      }
   }

   //
   // f0 -> [ "static" ]
   // f1 -> Block()
   //
   public void visit(Initializer n) {
      n.f1.accept(this);
   }

  /*
   //
   // f0 -> Literal()
   //       | ResultType() "." "class"
   //       | AllocationExpression()
   //       | "this"
   //       | "super"
   //       | "(" Expression() ")"
   //       | <IDENTIFIER>
   //
   public void visit(PrimaryPrefix n) {
      n.f0.accept(this);
   }

   //
   // f0 -> "." "this"
   //       | "." AllocationExpression()
   //       | "[" "[" DistArraySecRefDimList() "]" "]"
   //       | "[" TripletOrExpressionList() "]"
   //       | "." <IDENTIFIER>
   //       | Arguments()
   //
   public void visit(PrimarySuffix n) {
      n.f0.accept(this);
   }
  */

   //
   // f0 -> ( PrimitiveType() | Name() )
   // f1 -> ( ArraySignature() )*
   //
   public void visit(Type n) {
      TypeNode tl = new TypeNode(n);
      AbstractExpression rtn = tl;

      //      n.f0.accept(this);
      //      n.f1.accept(this);
   }

   //
   // f0 -> DistArraySecRefDim()
   // f1 -> ( "," DistArraySecRefDim() )*
   //
   public void visit(DistArraySecRefDimList n) {
      n.f0.accept(this);
      n.f1.accept(this);
   }

   //
   // f0 -> Triplet()
   //       | Expression()
   //       | KernelSubrange()
   //
   public void visit(DistArraySecRefDim n) {
      n.f0.accept(this);
   }

   //
   // f0 -> [ Expression() ]
   // f1 -> ":"
   // f2 -> [ Expression() ]
   // f3 -> [ ":" Expression() ]
   //
   public void visit(Triplet n) {
      n.f0.accept(this);
      n.f1.accept(this);
      n.f2.accept(this);
      n.f3.accept(this);
   }

   //
   // f0 -> "<" ( PrimaryExpression() | ":" ) ">"
   //       | "<" KernelSubrange() ">"
   //
   public void visit(KernelSubrange n) {
      n.f0.accept(this);
   }

   //
   // f0 -> TripletOrExpression()
   // f1 -> ( "," TripletOrExpression() )*
   //
   public void visit(TripletOrExpressionList n) {
      n.f0.accept(this);
      n.f1.accept(this);
   }

   //
   // f0 -> Triplet()
   //       | Expression()
   //
   public void visit(TripletOrExpression n) {
      n.f0.accept(this);
   }

   //
   // f0 -> <INTEGER_LITERAL>
   //       | <FLOATING_POINT_LITERAL>
   //       | <CHARACTER_LITERAL>
   //       | <STRING_LITERAL>
   //       | BooleanLiteral()
   //       | NullLiteral()
   //
   public void visit(Literal n) {
      n.f0.accept(this);
   }

   //
   // f0 -> "true"
   //       | "false"
   //
   public void visit(BooleanLiteral n) {
      n.f0.accept(this);
   }

   //
   // f0 -> "null"
   //
   public void visit(NullLiteral n) {
      n.f0.accept(this);
   }

   //
   // f0 -> "("
   // f1 -> [ ArgumentList() ]
   // f2 -> ")"
   //
   public void visit(Arguments n) {
      n.f1.accept(this);
   }

   //
   // f0 -> Expression()
   // f1 -> ( "," Expression() )*
   //
   public void visit(ArgumentList n) {
      n.f0.accept(this);
      n.f1.accept(this);
   }

   //
   // f0 -> "new" PrimitiveType() ArrayShapes()
   //       | "new" Name() ( ArrayShapes() | Arguments() [ ClassBody() ] )
   //
   public void visit(AllocationExpression n) {
      n.f0.accept(this);
   }

   //
   // f0 -> ( AllocatedShape() )+ ( ArraySignature() )* [ "on" GroupExpression() ]
   //       | ( "[" "]" )+ ArrayInitializer()
   //       | "#" [ "on" GroupExpression() ]
   //
   public void visit(ArrayShapes n) {
      n.f0.accept(this);
   }

   //
   // f0 -> "[" "[" Expression() ( "," Expression() )* "]" "]"
   //       | "[" Expression() "]"
   //
   public void visit(AllocatedShape n) {
      n.f0.accept(this);
   }

   //
   // f0 -> LabeledStatement()
   //       | Block()
   //       | EmptyStatement()
   //       | StatementExpression() ";"
   //       | SwitchStatement()
   //       | IfStatement()
   //       | WhileStatement()
   //       | DoStatement()
   //       | ForStatement()
   //       | BreakStatement()
   //       | ContinueStatement()
   //       | ReturnStatement()
   //       | ThrowStatement()
   //       | SynchronizedStatement()
   //       | TryStatement()
   //       | OnStatement()
   //       | AtStatement()
   //       | OverallStatement()
   //
   public void visit(Statement n) {
      if(n.f0.which == 3){
        NodeSequence seq = (NodeSequence)n.f0.choice;
        seq.elementAt(0).accept(this);
      }
      else{
        n.f0.accept(this);
      }
   }

   //
   // f0 -> <IDENTIFIER>
   // f1 -> ":"
   // f2 -> Statement()
   //
   public void visit(LabeledStatement n) {
      n.f2.accept(this);
   }

   //
   // f0 -> "{"
   // f1 -> ( BlockStatement() )*
   // f2 -> "}"
   //
   public void visit(Block n) {
      n.f1.accept(this);
   }

   //
   // f0 -> LocalVariableDeclaration() ";"
   //       | Statement()
   //       | UnmodifiedClassDeclaration()
   //       | UnmodifiedInterfaceDeclaration()
   //
   public void visit(BlockStatement n) {
      if(n.f0.which == 0) {
         NodeSequence seq = (NodeSequence)n.f0.choice;
         seq.elementAt(0).accept(this);
      }
      else{
         n.f0.accept(this);
      }
   }

   //
   // f0 -> [ "final" ]
   // f1 -> Type()
   // f2 -> VariableDeclarator()
   // f3 -> ( "," VariableDeclarator() )*
   //
   public void visit(LocalVariableDeclaration n) {
      n.f2.accept(this);

      if(n.f3.present()){
        for(int i=0; i < n.f3.size(); i++){
          NodeSequence seq = (NodeSequence)n.f3.elementAt(i);
          seq.elementAt(1).accept(this);
        }
      }
   }
   //
   // f0 -> "switch"
   // f1 -> "("
   // f2 -> Expression()
   // f3 -> ")"
   // f4 -> "{"
   // f5 -> ( SwitchLabel() ( BlockStatement() )* )*
   // f6 -> "}"
   //
   public void visit(SwitchStatement n) {
      n.f2.accept(this);
      n.f5.accept(this);
   }

   //
   // f0 -> "case" Expression() ":"
   //       | "default" ":"
   //
   public void visit(SwitchLabel n) {
      n.f0.accept(this);
   }

   //
   // f0 -> "if"
   // f1 -> "("
   // f2 -> Expression()
   // f3 -> ")"
   // f4 -> Statement()
   // f5 -> [ "else" Statement() ]
   //
   public void visit(IfStatement n) {
      n.f2.accept(this);
      n.f4.accept(this);
      n.f5.accept(this);
   }

   //
   // f0 -> "while"
   // f1 -> "("
   // f2 -> Expression()
   // f3 -> ")"
   // f4 -> Statement()
   //
   public void visit(WhileStatement n) {
      n.f2.accept(this);
      n.f4.accept(this);
   }

   //
   // f0 -> "do"
   // f1 -> Statement()
   // f2 -> "while"
   // f3 -> "("
   // f4 -> Expression()
   // f5 -> ")"
   // f6 -> ";"
   //
   public void visit(DoStatement n) {
      n.f1.accept(this);
      n.f4.accept(this);
   }

   //
   // f0 -> "for"
   // f1 -> "("
   // f2 -> [ ForInit() ]
   // f3 -> ";"
   // f4 -> [ Expression() ]
   // f5 -> ";"
   // f6 -> [ ForUpdate() ]
   // f7 -> ")"
   // f8 -> Statement()
   //
   public void visit(ForStatement n) {
      n.f2.accept(this);
      n.f4.accept(this);
      n.f6.accept(this);
      n.f8.accept(this);
   }

   //
   // f0 -> LocalVariableDeclaration()
   //       | StatementExpressionList()
   //
   public void visit(ForInit n) {
      n.f0.accept(this);
   }

   //
   // f0 -> StatementExpression()
   // f1 -> ( "," StatementExpression() )*
   //
   public void visit(StatementExpressionList n) {
      n.f0.accept(this);
      n.f1.accept(this);
   }

   //
   // f0 -> StatementExpressionList()
   //
   public void visit(ForUpdate n) {
      n.f0.accept(this);
   }

   //
   // f0 -> "throw"
   // f1 -> Expression()
   // f2 -> ";"
   //
   public void visit(ThrowStatement n) {
      n.f1.accept(this);
   }

   //
   // f0 -> "synchronized"
   // f1 -> "("
   // f2 -> Expression()
   // f3 -> ")"
   // f4 -> Block()
   //
   public void visit(SynchronizedStatement n) {
      n.f2.accept(this);
      n.f4.accept(this);
   }

   //
   // f0 -> "try"
   // f1 -> Block()
   // f2 -> ( "catch" "(" FormalParameter() ")" Block() )*
   // f3 -> [ "finally" Block() ]
   //
   public void visit(TryStatement n) {
      n.f1.accept(this);
      n.f2.accept(this);
      n.f3.accept(this);
   }

   //
   // f0 -> "on"
   // f1 -> "("
   // f2 -> Expression()
   // f3 -> ")"
   // f4 -> Statement()
   //
   public void visit(OnStatement n) {
      n.f2.accept(this);
      n.f4.accept(this);
   }

   // 
   // f0 -> "at" 
   // f1 -> "(" 
   // f2 -> NamedLocationList() 
   // f3 -> ")" 
   // f4 -> Statement() 
   // 
   public void visit(AtStatement n) {
      n.f2.accept(this);
      n.f4.accept(this);
   }

   /**
    * <PRE>
    * f0 -> &lt;IDENTIFIER&gt;
    * f1 -> "="
    * f2 -> PrimaryExpression()
    * </PRE>
    */
   public void visit(NamedLocation n) {
      n.f2.accept(this);
   }

   // 
   // f0 -> "overall" 
   // f1 -> "(" 
   // f2 -> NamedLocationRangeList() 
   // f3 -> ")" 
   // f4 -> Statement() 
   // 
   public void visit(OverallStatement n) {
      n.f2.accept(this);
      n.f4.accept(this);
   }

   /**
    * <PRE>
    * f0 -> &lt;IDENTIFIER&gt;
    * f1 -> "="
    * f2 -> PrimaryExpression()
    * f3 -> "for"
    * f4 -> Triplet()
    * </PRE>
    */
   public void visit(NamedLocationRange n) {
      n.f2.accept(this);
      n.f4.accept(this);
   }

}
