/*
 * File         : ClassReflectorManager.java
 * Author       : Guansong Zhang, Han-Ku Lee
 * Created      : Thu May 14 12:03:02 1998
 * Revision     : $Revision: 1.25 $
 * Updated      : $Date: 1999/12/01 19:06:52 $
 * Revised By   : Han-Ku Lee
 */

package hpjava.compiler.reflection;

import java.io.*;
import java.util.*;

import hpjava.compiler.util.*;

import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;

public class ClassReflectorManager {

  static final int TYPENUM = 64;
  static Hashtable nameTypeTable  = new Hashtable (TYPENUM);

  static final int FILENUM = 8;

  static ExportedClassNameManager expName;
  static LocalClassNameManager locName;

  static public ClassReflector lookup (ScalarReflectorDescriptor srd) {
    // for array, use ClassReflectorByDescriptor constructor

    String name = srd.name();

    ClassReflector reflector = (ClassReflector) nameTypeTable.get(name);

    //System.out.println("########### name: " + name);
    //System.out.println("########### nameTypeTable: " + nameTypeTable.toString());
    //System.out.println("########### reflector: " + reflector);

    if (reflector == null) {
      try {
        Class cls = Clazz.forName(name);

        reflector = new ClassReflectorByByteCode(cls);
      }
      catch(ClassNotFoundException e) {
        throw new Error("ScalarReflector created wrong. name: "+name);
        //System.out.println("ScalarReflector created wrong. name: "+name);
      }
    }
    
    return reflector;
  }

  static public ClassReflector lookup (String name) {
    // for array, use ClassReflectorByDescriptor constructor

    ClassReflector reflector = (ClassReflector) nameTypeTable.get(name);
    /*
    System.out.println();
    System.out.println("########### name: " + name);
    System.out.println("########### nameTypeTable: " + nameTypeTable.toString());
    System.out.println("########### reflector: " + reflector);
    System.out.println();
    */
    if (reflector == null) {
      try {

        //System.out.println(name);
        Class cls = Clazz.forName(name);

        reflector = new ClassReflectorByByteCode(cls);
      }
      catch(ClassNotFoundException e) {
        throw new Error("ScalarReflector created wrong. name: "+name);
        //System.out.println("ScalarReflector created wrong. name: "+name);        
      }
    }
    
    return reflector;
  }

  public ClassReflectorManager (Node root) {
    NameSpaceDescriptor nsd = NameSpaceManager.nameSpaceOf(root);

    ClassReflectorFinder finder =
        new ClassReflectorFinder(nsd.pname(), nsd.inames(), this.nameTypeTable);

    root.accept(finder);
  }

  public ClassReflectorManager (Project prj) {

    final int IMPORTPACKAGENUM=16;

    File srcPath,srcDirPath;
    String srcFilePathName, srcFileName;

    Enumeration enum = prj.rootFileTable.keys();

    expName = new ExportedClassNameManager(prj);
    locName = new LocalClassNameManager(prj);

    LOOP: while (enum.hasMoreElements()) {

      Node root = (Node) enum.nextElement();

      NameSpaceDescriptor nsd = NameSpaceManager.nameSpaceOf(root);

      ClassReflectorFinder finder =
        new ClassReflectorFinder(nsd.pname(), nsd.inames(), this.nameTypeTable);

      root.accept(finder);

    }

    if (prj.isVerbose()) {
			//System.err.println("Type table: "+nameTypeTable+" created.");
			//System.err.println("Type table created.");
			//System.err.println();

      Enumeration typeEnum = nameTypeTable.elements();
      while (typeEnum.hasMoreElements()) {
        ClassReflector cr = (ClassReflector) typeEnum.nextElement();
        cr.show(System.err);
      }

      System.err.println();
    }

  }

}

/*
 * File         : ClassReflectorFinder.java
 * Author       : Guansong Zhang
 * Created      : Thu May 14 12:03:02 1998
 * Revision     : $Revision: 1.25 $
 * Updated      : $Date: 1999/12/01 19:06:52 $
 * Revised By   : Han-Ku Lee
 */

class ClassReflectorFinder extends RecoveryVisitor {

  final int DEFAULT = 0;
  final int RECOVER = 1;
  final int LOCAL   = 2;

  private boolean debug=false;
  private IntStack status=new IntStack();

  private String pname;
  private Vector inames;

  private StringBuffer nameBuffer;

  protected Hashtable nameTypeTable;

  public ClassReflectorFinder(String pname, Vector inames) {
    nameTypeTable=new Hashtable ();

    this.pname = pname;
    this.inames = inames;

    nameBuffer = new StringBuffer(pname);

    status.push(DEFAULT);

    debug=true;
  }

  public ClassReflectorFinder(String pname, Vector inames, Hashtable nameTypeTable) {
    this.nameTypeTable=nameTypeTable;

    this.pname = pname;
    this.inames = inames;

    nameBuffer = new StringBuffer(pname);

    status.push(DEFAULT);

  }

   //
   // User-generated visitor methods below
   //

   //
   // f0 -> [ PackageDeclaration() ]
   // f1 -> ( ImportDeclaration() )*
   // f2 -> ( TypeDeclaration() )*
   // f3 -> < EOF >
   //
   public void visit(CompilationUnit n) {
     if ( n.f2.present() ) {
       // use RecoveryVisitor here;
       recoveryVisit(n.f2);
     }
     if (debug)
      System.out.println(nameTypeTable);
   }

   //
   // f0 -> ( "abstract" | "final" | "public" )*
   // f1 -> UnmodifiedClassDeclaration()
   //
   public void visit(ClassDeclaration n) {
     int orgLength = nameBuffer.length();

     if (nameBuffer.length()==0)
       nameBuffer.append(n.f1.f1.tokenImage);
     else
       nameBuffer.append("."+n.f1.f1.tokenImage);

     String classPackage = nameBuffer.toString()+".*";

     inames.addElement(classPackage);

     ClassReflectorBySourceCode cls = new ClassReflectorBySourceCode (pname, inames, n);
     nameTypeTable.put(cls.getName(), cls);

     // to find inner class;
     n.f1.f4.accept(this);

     inames.removeElement(classPackage);

     nameBuffer.setLength(orgLength);
   }

   //
   // f0 -> ( "static" | "abstract" | "final" | "public" | "protected" | "private" )*
   // f1 -> UnmodifiedClassDeclaration()
   //
   public void visit(NestedClassDeclaration n) {
     int orgLength = nameBuffer.length();

     if (nameBuffer.length()==0)
       nameBuffer.append(n.f1.f1.tokenImage);
     else
       nameBuffer.append("."+n.f1.f1.tokenImage);

     String classPackage = nameBuffer.toString()+".*";

     inames.addElement(classPackage);

     ClassReflectorBySourceCode cls = new ClassReflectorBySourceCode (pname, inames, n);
     nameTypeTable.put(cls.getName(), cls);

     // to find inner class;
     n.f1.f4.accept(this);

     inames.removeElement(classPackage);

     nameBuffer.setLength(orgLength);
   }


   // f0 -> ( "abstract" | "public" )*
   // f1 -> UnmodifiedInterfaceDeclaration()
   //
   public void visit(InterfaceDeclaration n) {
     int orgLength = nameBuffer.length();

     if (nameBuffer.length()==0)
       nameBuffer.append(n.f1.f1.tokenImage);
     else
       nameBuffer.append("."+n.f1.f1.tokenImage);

     String classPackage = nameBuffer.toString()+".*";

     inames.addElement(classPackage);

     ClassReflectorBySourceCode cls = new ClassReflectorBySourceCode (pname, inames, n);
     nameTypeTable.put(cls.getName(), cls);

     // to find inner class;
     n.f1.f4.accept(this);

     inames.removeElement(classPackage);

     nameBuffer.setLength(orgLength);
   }


   //
   // f0 -> ( "static" | "abstract" | "final" | "public" | "protected" | "private" )*
   // f1 -> UnmodifiedInterfaceDeclaration()
   //
   public void visit(NestedInterfaceDeclaration n) {
     int orgLength = nameBuffer.length();

     if (nameBuffer.length()==0)
       nameBuffer.append(n.f1.f1.tokenImage);
     else
       nameBuffer.append("."+n.f1.f1.tokenImage);

     String classPackage = nameBuffer.toString()+".*";

     inames.addElement(classPackage);

     ClassReflectorBySourceCode cls = new ClassReflectorBySourceCode (pname, inames, n);
     nameTypeTable.put(cls.getName(), cls);

     // to find inner class;
     n.f1.f4.accept(this);

     inames.removeElement(classPackage);

     nameBuffer.setLength(orgLength);
   }


  // only need to find local class.

   //
   // f0 -> "{"
   // f1 -> ( BlockStatement() )*
   // f2 -> "}"
   //
   public void visit(Block n) {
      status.push(LOCAL);
      n.f1.accept(this);
      status.pop();
   }

   //
   // f0 -> "class"
   // f1 -> <IDENTIFIER>
   // f2 -> [ "extends" Name() ]
   // f3 -> [ "implements" NameList() ]
   // f4 -> ClassBody()
   //
   public void visit(UnmodifiedClassDeclaration n) {

    // System.out.println("to construct a local");

     ClassReflectorBySourceCode cls = new ClassReflectorBySourceCode (pname, inames, n);
     nameTypeTable.put(cls.getName(), cls);

     // continue to find inner class;
      n.f4.accept(this);
   }

   //
   // f0 -> "interface"
   // f1 -> <IDENTIFIER>
   // f2 -> [ "extends" NameList() ]
   // f3 -> "{"
   // f4 -> ( InterfaceMemberDeclaration() )*
   // f5 -> "}"
   //
   public void visit(UnmodifiedInterfaceDeclaration n) {
     ClassReflectorBySourceCode cls = new ClassReflectorBySourceCode (pname, inames, n);
     nameTypeTable.put(cls.getName(), cls);

     // continue to find inner class;
      n.f4.accept(this);
   }

}

// The finder is responsible for finding inner class.
