/*
 * File         : LocalVariableManager.java
 * Author       : Guansong Zhang, Han-Ku Lee
 * Created      : Thu May 14 12:03:02 1998
 * Revision     : $Revision: 1.18 $
 * Updated      : $Date: 1999/12/01 19:06:55 $
 */

package hpjava.compiler.reflection;

import java.io.*;
import java.util.*;

import hpjava.compiler.util.*;

import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;

public class LocalVariableManager {

  static final int SYMBNUM = 128;  

  static Hashtable symbTypeTable  = new Hashtable (SYMBNUM);

  static public LocalVariableDescriptor lookup(NodeToken n)
    throws NoSuchNameException {
    String name = n.tokenImage;
    ScopeDescriptor scope = ScopeManager.scopeOf(n);

    LocalVariableDescriptor lvd; 

    while(scope != ScopeManager.globalScope) {
      LocalNameKey lnk = new LocalNameKey (name, scope);

      lvd = (LocalVariableDescriptor) symbTypeTable.get(lnk);
      
      if (lvd != null)
        return lvd;

      scope = ScopeManager.scopeOf(scope.node());
    }
    
    throw new NoSuchNameException (new PositionDescriptor(n));
  }

  static public void add(NodeToken n, Class cls){
    LocalNameKey lnk = new LocalNameKey (n);
     
    if (symbTypeTable.containsKey(lnk)) {
      String message = "variable "+n.tokenImage+" is already defined in this scope.";
      throw new CompilingError(message, n);
    }
    else {
      ClassTypeDescriptor ctd = new ClassTypeDescriptor(cls);
          
      LocalVariableDescriptor lvd = 
        new LocalVariableDescriptor(n.tokenImage, ctd);

      symbTypeTable.put(lnk, lvd);
    }
    
  }

  static public String typeSignatureOf(NodeToken n) 
    throws NoSuchNameException {
    return lookup(n).typeSignature();
  }


  static public ClassReflectorDescriptor classReflectorDescriptorOf(NodeToken n) 
    throws NoSuchNameException {
    return lookup(n).classReflectorDescriptor();
  }  

  public LocalVariableManager (Node root) {
    NameSpaceDescriptor nsd = NameSpaceManager.nameSpaceOf(root);

    LocalVariableFinder finder = 
      new LocalVariableFinder(nsd.pname(), nsd.inames(), this.symbTypeTable);

    root.accept(finder);   
  }

  public LocalVariableManager (Project prj) {
    
    Enumeration enum = prj.rootFileTable.keys();
   
    LOOP: while (enum.hasMoreElements()) {
      
      Node root = (Node) enum.nextElement();
      
      NameSpaceDescriptor nsd = NameSpaceManager.nameSpaceOf(root);
            
      LocalVariableFinder finder = 
        new LocalVariableFinder(nsd.pname(), nsd.inames(), this.symbTypeTable);

      root.accept(finder);   
    }
    
    if (prj.isVerbose()) {
      System.err.println("Symbol table: "+symbTypeTable+
                         " created with "+symbTypeTable.size()+" items.");
      System.err.println();
/*
      Enumeration typeEnum = nameTypeTable.elements();
      while (typeEnum.hasMoreElements()) {
      	ClassReflector cr = (ClassReflector) typeEnum.nextElement();
      	cr.show(System.err);
      }
        
      System.err.println();
*/      
      
    }       


  }

} 

class LocalVariableFinder extends RecoveryVisitor {

  final int DEFAULT = 0;
  final int RECOVER = 1;
  final int LOCAL   = 2;

  private boolean debug=false;
  private IntStack status=new IntStack();

  private String pname;
  private Vector inames;
  private StringBuffer nameBuffer;

  private ScalarReflectorDescriptor srd;

  private ArraySignaturesImage prefix;

  private boolean isFinal = false;
  private boolean isParameter = false;  

  protected Hashtable symbTypeTable;

  public LocalVariableFinder(String pname, Vector inames) {
    symbTypeTable=new Hashtable ();

    this.pname = pname;
    this.inames = inames;
    
    nameBuffer = new StringBuffer(pname);

    status.push(DEFAULT);

    debug=true;
  }

  public LocalVariableFinder(String pname, Vector inames, Hashtable symbTypeTable) {
    this.symbTypeTable=symbTypeTable;

    this.pname = pname;
    this.inames = inames;

    nameBuffer = new StringBuffer(pname);

    status.push(DEFAULT);

  }

  
   //
   // f0 -> "{"
   // f1 -> ( BlockStatement() )*
   // f2 -> "}"
   //
   public void visit(Block n) {
      recoveryVisit(n.f1);
   }
  
   //
   // f0 -> LocalVariableDeclaration() ";"
   //       | Statement()
   //       | UnmodifiedClassDeclaration()
   //       | UnmodifiedInterfaceDeclaration()
   //
   public void visit(BlockStatement n) {
      isParameter=false;
      n.f0.accept(this);
   }


   //
   // f0 -> [ "final" ]
   // f1 -> Type()
   // f2 -> VariableDeclarator()
   // f3 -> ( "," VariableDeclarator() )*
   //
   public void visit(LocalVariableDeclaration n) {
      if (n.f0.present()) 
        isFinal=true;
      else
        isFinal=false;

      n.f1.accept(this);
      
      n.f2.accept(this);
      if (n.f3.present())
        n.f3.accept(this);
   }

   //
   // f0 -> [ "final" ]
   // f1 -> Type()
   // f2 -> VariableDeclaratorId()
   //
   public void visit(FormalParameter n) {
      if (n.f0.present()) 
        isFinal=true;
      else
        isFinal=false;

      n.f1.accept(this);
      n.f2.accept(this);
   }

   //
   // f0 -> ( PrimitiveType() | Name() )
   // f1 -> ( ArraySignature() )*
   //
   public void visit(Type n) {
     switch (n.f0.which) { 
     case 0:
       this.srd = new PrimitiveTypeDescriptor((PrimitiveType)n.f0.choice);
       break;
     case 1:
       this.srd = new ClassTypeDescriptor(pname, inames, (Name)n.f0.choice); 
       break;  
     }

      prefix = new ArraySignaturesImage ((NodeListOptional)n.f1);
   }

   //
   // f0 -> VariableDeclaratorId()
   // f1 -> [ "=" VariableInitializer() ]
   //
   public void visit(VariableDeclarator n) {
      n.f0.accept(this);
      // should add this on for init value.
      n.f1.accept(this);
   }

   //
   // f0 -> <IDENTIFIER>
   // f1 -> ( ArraySignature() )*
   //
   public void visit(VariableDeclaratorId n) {

      LocalNameKey lnk = new LocalNameKey (n.f0);

      if (symbTypeTable.containsKey(lnk)) {
        String message = "variable "+n.f0.tokenImage+" is already defined in this scope.";
        throw new CompilingError(message, n.f0);
      }
      else {
        ArraySignaturesImage asd = this.prefix;
     
        if (n.f1.present()) {
          ArraySignaturesImage postfix =
            new ArraySignaturesImage((NodeListOptional)n.f1);
        
          if (asd !=null) 
            asd = new ArraySignaturesImage(postfix, this.prefix);
          else
            asd = postfix;  
        }
        
        ClassReflectorDescriptor crd;
      
        String signature;
      
        if (!asd.isEmpty()) {       
          crd  = new ArrayReflectorDescriptor (this.srd, asd);     
        }
        else
          crd = this.srd;
                
        LocalVariableDescriptor lvd = 
          new LocalVariableDescriptor(n.f0.tokenImage, crd, isFinal, isParameter);
      
        //symbTypeTable.put(lnk.toString(), lvd);  
        symbTypeTable.put(lnk, lvd);

      }
   }
   
   // cut out field declaration
   
   //
   // f0 -> ( "public" | "protected" | "private" | "static" | "final" | "transient" | "volatile" )*
   // f1 -> Type()
   // f2 -> VariableDeclarator()
   // f3 -> ( "," VariableDeclarator() )*
   // f4 -> ";"
   //
   public void visit(FieldDeclaration n) {

   }

   
   //
   // f0 -> "at"
   // f1 -> "("
   // f2 -> NamedLocationList()
   // f3 -> ")"
   // f4 -> Statement()
   //
   public void visit(AtStatement n) {
      //n.f0.accept(this);
      //n.f1.accept(this);
      n.f2.accept(this);
      //n.f3.accept(this);
      n.f4.accept(this);
   }

   //
   // f0 -> <IDENTIFIER>
   // f1 -> "="
   // f2 -> PrimaryExpression()
   //
   public void visit(NamedLocation n) {
     /*
     LocalNameKey lnk = new LocalNameKey (n.f0);
     
      if (symbTypeTable.containsKey(lnk)) {
        String message = "variable "+n.f0.tokenImage+" is already defined in this scope.";
        throw new CompilingError(message, n.f0);
      }
      else {
        try {
          Class cls = Class.forName("hpjava.lang.Location");
          
          ClassTypeDescriptor ctd = new ClassTypeDescriptor(cls);
          
          LocalVariableDescriptor lvd = 
            new LocalVariableDescriptor(n.f0.tokenImage, ctd);

          symbTypeTable.put(lnk, lvd);

        }
        catch (ClassNotFoundException e) {
          throw new Error();
        }
     }
     */

     try{
       LocalVariableManager.add(n.f0, Class.forName("hpjava.lang.Location"));
     }
     catch (ClassNotFoundException e) {
        throw new Error();
     }

   }

   //
   // f0 -> "overall"
   // f1 -> "("
   // f2 -> NamedLocationRangeList()
   // f3 -> ")"
   // f4 -> Statement()
   //
   public void visit(OverallStatement n) {

      n.f2.accept(this);

      n.f4.accept(this);
   }

   //
   // f0 -> <IDENTIFIER>
   // f1 -> "="
   // f2 -> PrimaryExpression()
   // f3 -> "for"
   // f4 -> Triplet()
   //
   public void visit(NamedLocationRange n) {
     /*
     LocalNameKey lnk = new LocalNameKey (n.f0);
     
      if (symbTypeTable.containsKey(lnk)) {
        String message = "variable "+n.f0.tokenImage+" is already defined in this scope.";
        throw new CompilingError(message, n.f0);
      }
      else {
        try {
          Class cls = Class.forName("hpjava.lang.Location");
          
          ClassTypeDescriptor ctd = new ClassTypeDescriptor(cls);
          
          LocalVariableDescriptor lvd = 
            new LocalVariableDescriptor(n.f0.tokenImage, ctd);

          symbTypeTable.put(lnk, lvd);

        }
        catch (ClassNotFoundException e) {
          throw new Error();
        }
      }
     */

     try{
       LocalVariableManager.add(n.f0, Class.forName("hpjava.lang.Location"));
     }
     catch (ClassNotFoundException e) {
        throw new Error();
     }
 
   }

}
