/*
 * File         : TreePrinter.java
 * Author       : Han-Ku Lee
 * Created      : Sun Jun  6 19:31:06 1999
 * Revision     : $Revision: 1.20 $
 * Updated      : $Date: 1999/10/25 22:14:59 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1999
 */

package hpjava.compiler.unparser;

//import misc.Spacing;
import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;

import java.util.*;
import java.io.*;

public class TreePrinter extends DepthFirstVisitor {
   protected PrintWriter out;
   protected Space spc ; 

  public TreePrinter() {
    out = new PrintWriter(System.out, true);
    spc = new Space(1); 
  }

  public TreePrinter(Writer o)  {
    out = new PrintWriter(o, true);
    spc = new Space(1); 
  }

  public TreePrinter(OutputStream o)  {
    out = new PrintWriter(o, true);
    spc = new Space(1); 
  }

   public void flushWriter()     { out.flush(); }

   //
   // Auto class visitors--probably don't need to be overridden.
   //

   public void visit(NodeToken n) {
      out.print(n.tokenImage + " ");
      flushWriter();
   }

   //
   // User-generated visitor methods below
   //

   //
   // f0 -> [ PackageDeclaration() ]
   // f1 -> ( ImportDeclaration() )*
   // f2 -> ( TypeDeclaration() )*
   // f3 -> < EOF >
   //
   public void visit(CompilationUnit n) {
      out.print(spc.spc);
      out.print("CompilationUnit");
      out.println();
      spc.updateSpc(+1);
      n.f0.accept(this);
      n.f1.accept(this);
      n.f2.accept(this);
      spc.updateSpc(-1);
      out.println();  /* if this is deleted, the last "}" disappears. */
   }

   //
   // f0 -> "package"
   // f1 -> Name()
   // f2 -> ";"
   //
   public void visit(PackageDeclaration n) {
      out.print(spc.spc);
      out.print("PackageDeclaration");
      out.println();
      spc.updateSpc(+1);
      out.print(spc.spc);
      n.f1.accept(this);
      out.println();
      spc.updateSpc(-1);
   }

   //
   // f0 -> "import"
   // f1 -> Name()
   // f2 -> [ "." "*" ]
   // f3 -> ";"
   //
   // uses the toString() method of the NodeTokens to print manually
   // (rather than visiting the tokens).
   //
   public void visit(ImportDeclaration n) {
      out.print(spc.spc);
      out.print("ImportDeclaration");
      out.println();
      spc.updateSpc(+1);
      out.print(spc.spc);
      n.f1.accept(this);
      if(n.f2.present()){
        n.f2.accept(this);
        out.println();
      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> ClassDeclaration()
   //       | InterfaceDeclaration()
   //       | ";"
   //
   public void visit(TypeDeclaration n) {
      out.print(spc.spc);
      out.print("TypeDeclaration");
      out.println();
      spc.updateSpc(+1);
      if(n.f0.which != 2) n.f0.accept(this);
      spc.updateSpc(-1);
   }

   //
   // f0 -> ( "abstract"| "final" | "public"  )*
   // f1 -> UnmodifiedClassDeclaration()
   //
   public void visit(ClassDeclaration n) {
      out.print(spc.spc);
      out.print("ClassDeclaration");
      out.println();
      spc.updateSpc(+1);
      if(n.f0.present()){
        out.print(spc.spc);
        n.f0.accept(this);
        out.println();
      }        
      n.f1.accept(this);
      spc.updateSpc(-1);
   }

   //
   // f0 -> "class"
   // f1 -> < IDENTIFIER >
   // f2 -> [ "extends" Name() ]
   // f3 -> [ "implements" NameList() ]
   // f4 -> ClassBody()
   //
   // continued from ClassDeclaration
   //
   public void visit(UnmodifiedClassDeclaration n) {
      out.print(spc.spc);
      out.print("UnmodifiedClassDeclaration");
      out.println();
      spc.updateSpc(+1);
      out.print(spc.spc);
      n.f0.accept(this);
      n.f1.accept(this);
      out.println();
      
      if(n.f2.present()){
        out.print(spc.spc);
        n.f2.accept(this);
        out.println();
      }

      if(n.f3.present()){
        out.print(spc.spc);
        n.f3.accept(this);
        out.println();
      }

      n.f4.accept(this);
      spc.updateSpc(-1);
   }

   //
   // f0 -> "{"
   // f1 -> ( ClassBodyDeclaration() )*
   // f2 -> "}"
   //
   public void visit(ClassBody n) {
      out.print(spc.spc);
      out.print("ClassBody");
      out.println();
      spc.updateSpc(+1);
      if(n.f1.present()) n.f1.accept(this);
      spc.updateSpc(-1);
   }

   //
   // f0 -> ("static"|"abstract"|"final"|"public"|"protected"|"private")*
   // f1 -> UnmodifiedClassDeclaration()
   //
   public void visit(NestedClassDeclaration n) {
      out.print(spc.spc);
      out.print("ClassDeclaration");
      out.println();
      spc.updateSpc(+1);
      if(n.f0.present()){
        out.print(spc.spc);
        n.f0.accept(this);
        out.println();
      }        
      n.f1.accept(this);
      spc.updateSpc(-1);
   }

   //
   // f0 -> Initializer()
   //       | NestedClassDeclaration()
   //       | NestedInterfaceDeclaration()
   //       | ConstructorDeclaration()
   //       | MethodDeclaration()
   //       | FieldDeclaration()
   //
   public void visit(ClassBodyDeclaration n) {
      out.print(spc.spc);
      out.print("ClassBodyDeclaration");
      out.println();
      spc.updateSpc(+1);
      n.f0.accept(this);
      spc.updateSpc(-1);
   }

   //
   // f0 -> ( "public"| "protected" | "private" | "static" | "abstract" | "final" | "native" | "synchronized"  )*
   // f1 -> ResultType()
   // f2 -> < IDENTIFIER >
   // f3 -> "("
   //
   // For lookahead only, so don't need to do this
   public void visit(MethodDeclarationLookahead n) {
      out.print(spc.spc);
      out.print("MethodDeclarationLookahead");
      out.println();
      spc.updateSpc(+1);
      if(n.f0.present()){
        out.print(spc.spc);
        n.f0.accept(this);
        out.println();
      }        
      n.f1.accept(this);
      out.print(spc.spc);
      n.f2.accept(this);
      out.println();
      spc.updateSpc(-1);      // #####Do I have to put this on?#####
   }

   //
   // f0 -> ( "abstract"| "public"  )*
   // f1 -> UnmodifiedInterfaceDeclaration()
   //
   public void visit(InterfaceDeclaration n) {
      out.print(spc.spc);
      out.print("InterfaceDeclaration");
      out.println();
      spc.updateSpc(+1);
      if(n.f0.present()){
        out.print(spc.spc);
        n.f0.accept(this);
        out.println();
      }        
      n.f1.accept(this);
      spc.updateSpc(-1);
   }

   //
   // f0 -> ("static"|"abstract"|"final"|"public"|"protected"|"private")*
   // f1 -> UnmodifiedInterfaceDeclaration()
   //
   public void visit(NestedInterfaceDeclaration n) {
      out.print(spc.spc);
      out.print("NestedInterfaceDeclaration");
      out.println();
      spc.updateSpc(+1);
      if(n.f0.present()){
        out.print(spc.spc);
        n.f0.accept(this);
        out.println();
      }        
      n.f1.accept(this);
      spc.updateSpc(-1);
   }

   //
   // f0 -> "interface"
   // f1 -> < IDENTIFIER >
   // f2 -> [ "extends" NameList() ]
   // f3 -> "{"
   // f4 -> ( InterfaceMemberDeclaration() )*
   // f5 -> "}"
   //
   public void visit(UnmodifiedInterfaceDeclaration n) {
      out.print(spc.spc);
      out.print("UnmodifiedInterfaceDeclaration");
      out.println();
      spc.updateSpc(+1);
      out.print(spc.spc);
      n.f0.accept(this);
      n.f1.accept(this);
      out.println();
      if(n.f2.present()){
        out.print(spc.spc);
        n.f2.accept(this);
        out.println();
      }
      n.f4.accept(this);
      spc.updateSpc(-1);

   }

   //
   // f0 -> NestedClassDeclaration()
   //       | NestedInterfaceDeclaration()
   //       | MethodDeclaration()
   //       | FieldDeclaration()
   //
   public void visit(InterfaceMemberDeclaration n) {
      out.print(spc.spc);
      out.print("InterfaceMemberDeclaration");
      out.println();
      spc.updateSpc(+1);
      n.f0.accept(this);
      spc.updateSpc(-1);
   }

   //
   // f0 -> ("public"|"protected"|"private"|"static"|"final"|"transient"|"volatile")*
   // f1 -> Type()
   // f2 -> VariableDeclarator()
   // f3 -> ( "," VariableDeclarator() )*
   // f4 -> ";"
   //
   public void visit(FieldDeclaration n) {
      out.print(spc.spc);
      out.print("FieldDeclaration");
      out.println();
      spc.updateSpc(+1);
      if(n.f0.present()){
        out.print(spc.spc);
        n.f0.accept(this);
        out.println();
      }
      n.f1.accept(this);
      n.f2.accept(this);
      if(n.f3.present()){
        for(int i=0; i<n.f3.size(); i++){
          NodeSequence seq = (NodeSequence)n.f3.elementAt(i);
          seq.elementAt(1).accept(this);
        } 
      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> VariableDeclaratorId()
   // f1 -> [ "=" VariableInitializer() ]
   //
   public void visit(VariableDeclarator n) {
      out.print(spc.spc);
      out.print("VariableDeclarator");
      out.println();
      spc.updateSpc(+1);
      n.f0.accept(this);
      if(n.f1.present()){
        NodeSequence seq = (NodeSequence)n.f1.node;
        out.print(spc.spc);
        seq.elementAt(0).accept(this); 
        out.println();
        seq.elementAt(1).accept(this);  
      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> <IDENTIFIER>
   // f1 -> ( ArraySignature() )*
   //
   public void visit(VariableDeclaratorId n) {
      out.print(spc.spc);
      out.print("VariableDeclaratorId");
      out.println();
      spc.updateSpc(+1);
      out.print(spc.spc);
      n.f0.accept(this);
      out.println();
      if(n.f1.present()){
        n.f1.accept(this);
      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> ArrayInitializer()
   //       | Expression()
   //
   public void visit(VariableInitializer n) {
      out.print(spc.spc);
      out.print("VariableInitializer");
      out.println();
      spc.updateSpc(+1);
      n.f0.accept(this);
      spc.updateSpc(-1);
   }

   //
   // f0 -> "{"
   // f1 -> [ VariableInitializer() ( "," VariableInitializer() )* ]
   // f2 -> [ "," ]
   // f3 -> "}"
   //
   public void visit(ArrayInitializer n) {
      out.print(spc.spc);
      out.print("ArrayInitializer");
      out.println();
      spc.updateSpc(+1);
      out.print(spc.spc);
      n.f0.accept(this);
      out.println();

      if(n.f1.present()){
        NodeSequence seq1 = (NodeSequence)n.f1.node;
        seq1.elementAt(0).accept(this);
 
        if(((NodeListOptional)seq1.elementAt(1)).present()){
          NodeListOptional nlo = (NodeListOptional)seq1.elementAt(1);
          for(int j=0; j<nlo.size(); j++){
            NodeSequence seq2 = (NodeSequence)nlo.elementAt(j);
            seq2.elementAt(1).accept(this);
          }
        }

      }

      out.print(spc.spc);
      n.f3.accept(this);
      out.println();

      spc.updateSpc(-1);
   }

   //
   // f0 -> ( "public"| "protected" | "private" | "static" | "abstract" | "final" | "native" | "synchronized"  )*
   // f1 -> ResultType()
   // f2 -> MethodDeclarator()
   // f3 -> [ "throws" NameList() ]
   // f4 -> ( Block()| ";"  )
   //
   public void visit(MethodDeclaration n) {
      out.print(spc.spc);
      out.print("MethodDeclaration");
      out.println();
      spc.updateSpc(+1);
      if(n.f0.present()){
        out.print(spc.spc);
        n.f0.accept(this);
        out.println();
      }
      n.f1.accept(this);
      n.f2.accept(this);
      if(n.f3.present()){
        out.print(spc.spc);
        n.f3.accept(this);
        out.println();
      }
      if(n.f4.which == 0) n.f4.accept(this);
      spc.updateSpc(-1);
   }

   //
   // f0 -> <IDENTIFIER>
   // f1 -> FormalParameters()
   // f2 -> ( ArraySignature() )*
   //
   public void visit(MethodDeclarator n) {
      out.print(spc.spc);
      out.print("MethodDeclarator");
      out.println();
      spc.updateSpc(+1);
      out.print(spc.spc);
      n.f0.accept(this);
      out.println();
      n.f1.accept(this);
      n.f2.accept(this);
      spc.updateSpc(-1);
   }

   //
   // f0 -> "("
   // f1 -> [ FormalParameter() ( "," FormalParameter() )* ]
   // f2 -> ")"
   //
   public void visit(FormalParameters n) {
      out.print(spc.spc);
      out.print("FormalParameters");
      out.println();
      spc.updateSpc(+1);

      if(n.f1.present()){
        NodeSequence seq1 = (NodeSequence)n.f1.node;
        seq1.elementAt(0).accept(this);
 
        if(((NodeListOptional)seq1.elementAt(1)).present()){
          NodeListOptional nlo = (NodeListOptional)seq1.elementAt(1);
          for(int j=0; j<nlo.size(); j++){
            NodeSequence seq2 = (NodeSequence)nlo.elementAt(j);
            seq2.elementAt(1).accept(this);
          }
        }

      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> [ "final" ]
   // f1 -> Type()
   // f2 -> VariableDeclaratorId()
   //
   public void visit(FormalParameter n) {
      out.print(spc.spc);
      out.print("FormalParameter");
      out.println();
      spc.updateSpc(+1);
      if(n.f0.present()){
        out.print(spc.spc);
        n.f0.accept(this);
        out.println();
      }
      n.f1.accept(this);
      n.f2.accept(this);
      spc.updateSpc(-1);
   }

   //
   // f0 -> [ "public"| "protected" | "private"  ]
   // f1 -> < IDENTIFIER >
   // f2 -> FormalParameters()
   // f3 -> [ "throws" NameList() ]
   // f4 -> "{"
   // f5 -> [ ExplicitConstructorInvocation() ]
   // f6 -> ( BlockStatement() )*
   // f7 -> "}"
   //
   public void visit(ConstructorDeclaration n) {
      out.print(spc.spc);
      out.print("ConstructorDeclaration");
      out.println();
      spc.updateSpc(+1);

      if(n.f0.present()){
        out.print(spc.spc);
        n.f0.accept(this);
        out.println();
      }

      out.print(spc.spc);
      n.f1.accept(this);
      out.println();
      n.f2.accept(this);

      if(n.f3.present()){
        out.print(spc.spc);
        n.f3.accept(this);
        out.println();
      } 
      if(n.f5.present()){
        n.f5.accept(this);
      }
      if(n.f6.present()){
        n.f6.accept(this);
      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> "this" Arguments() ";"
   //       | [ PrimaryExpression() "." ] "super" Arguments() ";"
   //
   public void visit(ExplicitConstructorInvocation n) {
      //
      // This code may not be appropriate if "which" is removed from NodeChoice
      //
      out.print(spc.spc);
      out.print("ExplicitConstructorInvocation");
      out.println();
      spc.updateSpc(+1);
      if ( n.f0.which == 0 ) {

        NodeSequence seq = (NodeSequence)n.f0.choice; 
        out.print(spc.spc);
        seq.elementAt(0).accept(this);
        out.println();
        seq.elementAt(1).accept(this);
      }
      else {
        NodeSequence seq = (NodeSequence)n.f0.choice;
        if(seq.elementAt(0) != null){
          NodeSequence seq2 = (NodeSequence)seq.elementAt(0); 
          seq2.elementAt(0).accept(this);
          out.print(spc.spc);
          seq2.elementAt(1).accept(this);
          out.println();
        }
        out.print(spc.spc);
        seq.elementAt(1).accept(this);
        out.println();
        seq.elementAt(2).accept(this);
      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> [ "static" ]
   // f1 -> Block()
   //
   public void visit(Initializer n) {
      out.print(spc.spc);
      out.print("Initializer");
      out.println();
      spc.updateSpc(+1);
      out.println();
      out.print(spc.spc);
      n.f0.accept(this);
      n.f1.accept(this);
      spc.updateSpc(-1);
   }

   //
   // f0 -> "[" "]"
   //       | "[" "[" DistArrayDimAttList() "]" "]"
   //       | "#"
   //
   public void visit(ArraySignature n) {
      out.print(spc.spc);
      out.print("ArraySignature");
      out.println();
      spc.updateSpc(+1);
      out.print(spc.spc);

      if(n.f0.which == 1){
        NodeSequence seq = (NodeSequence)n.f0.choice;
        seq.elementAt(0).accept(this);
        seq.elementAt(1).accept(this);
        out.println();
        seq.elementAt(2).accept(this);
        out.print(spc.spc);
        seq.elementAt(3).accept(this);
        seq.elementAt(4).accept(this);              
      }
      else{
        n.f0.accept(this);
      }
      out.println();

      spc.updateSpc(-1);
   }

   //
   // f0 -> [ "*" ]
   // f1 -> ( "," [ "*" ] )*
   //
   public void visit(DistArrayDimAttList n) {
      out.print(spc.spc);
      out.print("DistArrayDimAttList");
      out.println();
      spc.updateSpc(+1);
      if(n.f0.present()){
        out.print(spc.spc);
        n.f0.accept(this);
        out.println();
      }
      if(n.f1.present()){
        for(int i=0; i<n.f1.size(); i++){
          NodeSequence seq = (NodeSequence)n.f1.elementAt(i);
          out.print(spc.spc);
          seq.elementAt(0).accept(this);
          out.println();
          if(seq.elementAt(1) != null){
            out.print(spc.spc);
            seq.elementAt(1).accept(this);
            out.println();
          }       
        } 
      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> DistArraySecRefDim()
   // f1 -> ( "," DistArraySecRefDim() )*
   //
   public void visit(DistArraySecRefDimList n) {
      out.print(spc.spc);
      out.print("DistArraySecRefDimList");
      out.println();

      spc.updateSpc(+1);

      n.f0.accept(this);

      if(n.f1.present()){
        for(int i=0; i<n.f1.size(); i++){
          NodeSequence seq = (NodeSequence)n.f1.elementAt(i);
          seq.elementAt(1).accept(this);
        } 
      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> Triplet()
   //       | Expression()
   //       | KernelSubrange()
   //
   public void visit(DistArraySecRefDim n) {
      out.print(spc.spc);
      out.print("DistArraySecRefDim");
      out.println();
      spc.updateSpc(+1);
      n.f0.accept(this);
      spc.updateSpc(-1);
   }

   //
   // f0 -> [ Expression() ]
   // f1 -> ":"
   // f2 -> [ Expression() ]
   // f3 -> [ ":" Expression() ]
   //
   public void visit(Triplet n) {
      out.print(spc.spc);
      out.print("Triplet");
      out.println();
      spc.updateSpc(+1);
      if(n.f0.present()) n.f0.accept(this);
      out.print(spc.spc);      
      n.f1.accept(this);
      out.println();
      if(n.f2.present()) n.f2.accept(this);
      if(n.f3.present()){
        NodeSequence seq = (NodeSequence)n.f3.node;
        out.print(spc.spc);
        seq.elementAt(0).accept(this);
        out.println();
        seq.elementAt(1).accept(this); 
      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> "<" ( PrimaryExpression() | ":" ) ">"
   //       | "<" KernelSubrange() ">"
   //
   public void visit(KernelSubrange n) {
      out.print(spc.spc);
      out.print("KernelSubrange");
      out.println();
      spc.updateSpc(+1);

      NodeSequence seq = (NodeSequence)n.f0.choice;

      if(n.f0.which == 0){
        out.print(spc.spc);
        seq.elementAt(0).accept(this);        
        out.println();
        if(seq.elementAt(1) != null){
          NodeChoice nc = (NodeChoice)seq.elementAt(1);
          if(nc.which == 0){
            nc.accept(this);
          }
          else{
            out.print(spc.spc);
            nc.accept(this);
            out.println();
          }
        }
        out.print(spc.spc);
        seq.elementAt(2).accept(this);        
        out.println(); 
      }
      else{
        out.print(spc.spc);
        seq.elementAt(0).accept(this);        
        out.println();
        seq.elementAt(1).accept(this);
        out.print(spc.spc);
        seq.elementAt(2).accept(this);
        out.println();
      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> Triplet()
   //       | Expression()
   //
   public void visit(TripletOrExpression n) {
      out.print(spc.spc);
      out.print("TripletOrExpression");
      out.println();
      spc.updateSpc(+1);
      n.f0.accept(this);
      spc.updateSpc(-1);
   }

   //
   // f0 -> TripletOrExpression()
   // f1 -> ( "," TripletOrExpression() )*
   //
   public void visit(TripletOrExpressionList n) {
      out.print(spc.spc);
      out.print("TripletOrExpressionList");
      out.println();
      spc.updateSpc(+1);
      n.f0.accept(this);
      if(n.f1.present()){
        for(int i=0; i<n.f1.size(); i++){
          NodeSequence seq = (NodeSequence)n.f1.elementAt(i);
          seq.elementAt(1).accept(this);
        } 
      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> ( PrimitiveType() | Name() )
   // f1 -> ( ArraySignature() )*
   //
   public void visit(Type n) {
      out.print(spc.spc);
      out.print("Type");
      out.println();
      spc.updateSpc(+1);
      n.f0.accept(this);
      if(n.f0.which == 1){
        out.println();
      }

      if(n.f1.present()){
        n.f1.accept(this);
      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> "boolean"
   //       | "char"
   //       | "byte"
   //       | "short"
   //       | "int"
   //       | "long"
   //       | "float"
   //       | "double"
   //
   public void visit(PrimitiveType n) {
      out.print(spc.spc);
      out.print("PrimitiveType");
      out.println();
      spc.updateSpc(+1);
      out.print(spc.spc);
      n.f0.accept(this);
      out.println();
      spc.updateSpc(-1);
   }

   //
   // f0 -> "public void"
   //       | Type()
   //
   public void visit(ResultType n) {
      out.print(spc.spc);
      out.print("ResultType ");
      out.println();
      spc.updateSpc(+1);
      out.print(spc.spc);
      n.f0.accept(this);
      if(n.f0.which == 0){
       out.println();
      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> < IDENTIFIER >
   // f1 -> ( "." < IDENTIFIER >  )*
   //
   public void visit(Name n) {
      out.print(spc.spc);
      n.f0.accept(this);
      n.f1.accept(this);
   }

   //
   // f0 -> Name()
   // f1 -> ( "," Name() )*
   //
   // Add spaces?
   //
   public void visit(NameList n) {
      out.print(spc.spc);
      n.f0.accept(this);
      n.f1.accept(this);
   }

   //
   // f0 -> ConditionalExpression()
   // f1 -> [ AssignmentOperator() Expression() ]
   //
   public void visit(Expression n) {
      out.print(spc.spc);
      out.print("Expression");
      out.println();
      spc.updateSpc(+1);
      n.f0.accept(this);
      if(n.f1.present()) n.f1.accept(this);
      spc.updateSpc(-1);
   }

   //
   // f0 -> "="
   //       | "*="
   //       | "/="
   //       | "%="
   //       | "+="
   //       | "-="
   //       | "<<="
   //       | ">>="
   //       | ">>>="
   //       | "&="
   //       | "^="
   //       | "|="
   //
   public void visit(AssignmentOperator n) {
      out.print(spc.spc);
      out.print("AssignmentOperator");
      out.println();
      spc.updateSpc(+1);
      out.print(spc.spc);
      n.f0.accept(this);
      out.println();
      spc.updateSpc(-1);
   }

   //
   // f0 -> ConditionalOrExpression()
   // f1 -> [ "?" Expression() ":" ConditionalExpression() ]
   //
   public void visit(ConditionalExpression n) {
      out.print(spc.spc);
      out.print("ConditionalExpression");
      out.println();
      spc.updateSpc(+1);
      n.f0.accept(this);      

      if(n.f1.present()){
        NodeSequence seq = (NodeSequence)n.f1.node;
        out.print(spc.spc);
        seq.elementAt(0);
        out.println();
        seq.elementAt(1);
        out.print(spc.spc);
        seq.elementAt(2);
        out.println();
        seq.elementAt(3); 
      }

      spc.updateSpc(-1);
   }

   //
   // f0 -> ConditionalAndExpression()
   // f1 -> ( "||" ConditionalAndExpression() )*
   //
   public void visit(ConditionalOrExpression n) {
      out.print(spc.spc);
      out.print("ConditionalOrExpression");
      out.println();
      spc.updateSpc(+1);
      n.f0.accept(this);
      if(n.f1.present()){
        for(int i=0; i<n.f1.size(); i++){
          NodeSequence seq = (NodeSequence)n.f1.elementAt(i);
          out.print(spc.spc);
          seq.elementAt(0).accept(this);
          out.println();
          seq.elementAt(1).accept(this);
        } 

      }
      spc.updateSpc(-1);

   }

   //
   // f0 -> InclusiveOrExpression()
   // f1 -> ( "&&" InclusiveOrExpression() )*
   //
   public void visit(ConditionalAndExpression n) {
      out.print(spc.spc);
      out.print("ConditionalAndExpression");
      out.println();
      spc.updateSpc(+1);
      n.f0.accept(this);
      if(n.f1.present()){
        for(int i=0; i<n.f1.size(); i++){
          NodeSequence seq = (NodeSequence)n.f1.elementAt(i);
          out.print(spc.spc);
          seq.elementAt(0).accept(this);
          out.println();
          seq.elementAt(1).accept(this);
        } 

      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> ExclusiveOrExpression()
   // f1 -> ( "|" ExclusiveOrExpression() )*
   //
   public void visit(InclusiveOrExpression n) {
      out.print(spc.spc);
      out.print("InclusiveOrExpression");
      out.println();
      spc.updateSpc(+1);
      n.f0.accept(this);
      if(n.f1.present()){
        for(int i=0; i<n.f1.size(); i++){
          NodeSequence seq = (NodeSequence)n.f1.elementAt(i);
          out.print(spc.spc);
          seq.elementAt(0).accept(this);
          out.println();
          seq.elementAt(1).accept(this);
        } 

      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> AndExpression()
   // f1 -> ( "^" AndExpression() )*
   //
   public void visit(ExclusiveOrExpression n) {
      out.print(spc.spc);
      out.print("ExclusiveOrExpression");
      out.println();
      spc.updateSpc(+1);
      n.f0.accept(this);
      if(n.f1.present()){
        for(int i=0; i<n.f1.size(); i++){
          NodeSequence seq = (NodeSequence)n.f1.elementAt(i);
          out.print(spc.spc);
          seq.elementAt(0).accept(this);
          out.println();
          seq.elementAt(1).accept(this);
        } 

      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> EqualityExpression()
   // f1 -> ( "&" EqualityExpression() )*
   //
   public void visit(AndExpression n) {
      out.print(spc.spc);
      out.print("AndExpression");
      out.println();
      spc.updateSpc(+1);
      n.f0.accept(this);
      if(n.f1.present()){
        for(int i=0; i<n.f1.size(); i++){
          NodeSequence seq = (NodeSequence)n.f1.elementAt(i);
          out.print(spc.spc);
          seq.elementAt(0).accept(this);
          out.println();
          seq.elementAt(1).accept(this);
        } 

      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> InstanceOfExpression()
   // f1 -> ( ( "=="| "!="  ) InstanceOfExpression() )*
   //
   public void visit(EqualityExpression n) {
      out.print(spc.spc);
      out.print("EqualityExpression");
      out.println();
      spc.updateSpc(+1);
      n.f0.accept(this);
      if(n.f1.present()){
        for(int i=0; i<n.f1.size(); i++){
          NodeSequence seq = (NodeSequence)n.f1.elementAt(i);
          out.print(spc.spc);
          seq.elementAt(0).accept(this);
          out.println();
          seq.elementAt(1).accept(this);
        } 

      }
      spc.updateSpc(-1);

   }

   //
   // f0 -> RelationalExpression()
   // f1 -> [ "instanceof" Type() ]
   //
   public void visit(InstanceOfExpression n) {
      out.print(spc.spc);
      out.print("InstanceOfExpression");
      out.println();
      spc.updateSpc(+1);
      n.f0.accept(this);
      if(n.f1.present()){
        NodeSequence seq = (NodeSequence)n.f1.node;
        out.print(spc.spc);
        seq.elementAt(0).accept(this);
        out.println();
        seq.elementAt(1).accept(this);
        
      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> ShiftExpression()
   // f1 -> ( ( "<"| ">" | "<=" | ">="  ) ShiftExpression() )*
   //
   public void visit(RelationalExpression n) {
      out.print(spc.spc);
      out.print("RelationalExpression");
      out.println();
      spc.updateSpc(+1);
      n.f0.accept(this);
      if(n.f1.present()){
        for(int i=0; i<n.f1.size(); i++){
          NodeSequence seq = (NodeSequence)n.f1.elementAt(i);
          out.print(spc.spc);
          seq.elementAt(0).accept(this);
          out.println();
          seq.elementAt(1).accept(this);
        } 

      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> AdditiveExpression()
   // f1 -> ( ( "<<"| ">>" | ">>>"  ) AdditiveExpression() )*
   //
   public void visit(ShiftExpression n) {
      out.print(spc.spc);
      out.print("ShiftExpression");
      out.println();
      spc.updateSpc(+1);
      n.f0.accept(this);
      if(n.f1.present()){
        for(int i=0; i<n.f1.size(); i++){
          NodeSequence seq = (NodeSequence)n.f1.elementAt(i);
          out.print(spc.spc);
          seq.elementAt(0).accept(this);
          out.println();
          seq.elementAt(1).accept(this);
        } 

      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> MultiplicativeExpression()
   // f1 -> ( ( "+"| "-"  ) MultiplicativeExpression() )*
   //
   public void visit(AdditiveExpression n) {
      out.print(spc.spc);
      out.print("AdditiveExpression");
      out.println();
      spc.updateSpc(+1);
      n.f0.accept(this);
      if(n.f1.present()){
        for(int i=0; i<n.f1.size(); i++){
          NodeSequence seq = (NodeSequence)n.f1.elementAt(i);
          out.print(spc.spc);
          seq.elementAt(0).accept(this);
          out.println();
          seq.elementAt(1).accept(this);
        } 

      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> UnaryExpression()
   // f1 -> ( ( "*"| "/" | "%"  ) UnaryExpression() )*
   //
   public void visit(MultiplicativeExpression n) {
      out.print(spc.spc);
      out.print("MultiplicativeExpression");
      out.println();
      spc.updateSpc(+1);
      n.f0.accept(this);
      if(n.f1.present()){
        for(int i=0; i<n.f1.size(); i++){
          NodeSequence seq = (NodeSequence)n.f1.elementAt(i);
          out.print(spc.spc);
          seq.elementAt(0).accept(this);
          out.println();
          seq.elementAt(1).accept(this);
        } 

      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> ( "+"| "-"  ) UnaryExpression()
   //       | PreIncrementExpression()
   //       | PreDecrementExpression()
   //       | UnaryExpressionNotPlusMinus()
   //
   public void visit(UnaryExpression n) {
      out.print(spc.spc);
      out.print("UnaryExpression");
      out.println();
      spc.updateSpc(+1);
      if(n.f0.which == 0){
        NodeSequence seq = (NodeSequence)n.f0.choice;
        out.print(spc.spc); 
        seq.elementAt(0).accept(this);
        out.println();
        seq.elementAt(1).accept(this);
      }
      else{
        n.f0.accept(this);
      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> "++"
   // f1 -> PrimaryExpression()
   //
   public void visit(PreIncrementExpression n) {
      out.print(spc.spc);
      out.print("PreIncrementExpression");
      out.println();
      spc.updateSpc(+1);
      out.print(spc.spc);
      n.f0.accept(this);
      out.println();
      n.f1.accept(this);
      spc.updateSpc(-1);
   }

   //
   // f0 -> "--"
   // f1 -> PrimaryExpression()
   //
   public void visit(PreDecrementExpression n) {
      out.print(spc.spc);
      out.print("PreDecrementExpression");
      out.println();
      spc.updateSpc(+1);
      out.print(spc.spc);
      n.f0.accept(this);
      out.println();
      n.f1.accept(this);
      spc.updateSpc(-1);
   }

   //
   // f0 -> ( "~"| "!"  ) UnaryExpression()
   //       | CastExpression()
   //       | PostfixExpression()
   //
   public void visit(UnaryExpressionNotPlusMinus n) {
      out.print(spc.spc);
      out.print("UnaryExpressionNotPlusMinus");
      out.println();
      spc.updateSpc(+1);
      if(n.f0.which == 0){
        NodeSequence seq = (NodeSequence)n.f0.choice;
        out.print(spc.spc); 
        seq.elementAt(0).accept(this);
        out.println();
        seq.elementAt(1).accept(this);
      }
      else{
        n.f0.accept(this);
      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> "(" PrimitiveType()
   //       | "(" Name() ArraySignature()
   //       | "(" Name() ")" ( "~" | "!" | "(" | < IDENTIFIER >  | "this" | "super" | "new" | Literal()  )
   //
   public void visit(CastLookahead n) {
      out.print(spc.spc);
      out.print("CastLookahead");
      out.println();
      spc.updateSpc(+1);

      NodeSequence seq = (NodeSequence)n.f0.choice;
      if(n.f0.which == 0){
        out.print(spc.spc);
        seq.elementAt(0).accept(this);      
        out.println();
        seq.elementAt(1).accept(this);
      }
      else if(n.f0.which == 1){
        out.print(spc.spc);
        seq.elementAt(0).accept(this);
        out.println();
        seq.elementAt(1).accept(this);
        out.println();
        seq.elementAt(2).accept(this);
      }
      else{
        out.print(spc.spc);
        seq.elementAt(0).accept(this);
        out.println();
        seq.elementAt(1).accept(this);
        out.println();
        out.print(spc.spc);
        seq.elementAt(2).accept(this);
        out.println();
        if(((NodeChoice)seq.elementAt(3)).which == 7){
          seq.elementAt(3).accept(this); 
        }
        else{
         out.print(spc.spc); 
         seq.elementAt(3).accept(this);
         out.println();
        }
      
      }
      
      spc.updateSpc(-1);
   }

   //
   // f0 -> PrimaryExpression()
   // f1 -> [ "++" | "--" | "`" ]
   //
   public void visit(PostfixExpression n) {
      out.print(spc.spc);
      out.print("PostfixExpression");
      out.println();
      spc.updateSpc(+1);
      n.f0.accept(this);
      //if(n.f1.present()){  
      //  out.print(spc.spc);
      //  n.f1.accept(this);
      //  out.println();
      //}

      if(n.f1.present()){  
        out.print(spc.spc);

        
        NodeChoice nc = (NodeChoice)n.f1.node;
        if(nc.which == 2){
          out.print("Index_of_Location");
          out.println();
          out.print(spc.spc);
          n.f1.accept(this);
          out.println();
        }
        else
          n.f1.accept(this);

        out.println();
      }

      spc.updateSpc(-1);
   }

   //
   // f0 -> "(" Type() ")" UnaryExpression()
   //       | "(" Type() ")" UnaryExpressionNotPlusMinus()
   //
   public void visit(CastExpression n) {
      out.print(spc.spc);
      out.print("CastExpression");
      out.println();
      spc.updateSpc(+1);
      n.f0.accept(this);
      spc.updateSpc(-1);
   }

   //
   // f0 -> PrimaryPrefix()
   // f1 -> ( PrimarySuffix() )*
   //
   public void visit(PrimaryExpression n) {
      out.print(spc.spc);
      out.print("PrimaryExpression");
      out.println();
      spc.updateSpc(+1);
      n.f0.accept(this);
      n.f1.accept(this);
      spc.updateSpc(-1);
   }

   //
   // f0 -> Literal()
   //       | ResultType() "." "class"
   //       | AllocationExpression()
   //       | "this"
   //       | "super"
   //       | "(" Expression() ")"
   //       | <IDENTIFIER>
   //
   public void visit(PrimaryPrefix n) {
      out.print(spc.spc);
      out.print("PrimaryPrefix ");
      out.println();
      spc.updateSpc(+1);

      if(n.f0.which == 5){
        NodeSequence seq = (NodeSequence)n.f0.choice;
        out.print(spc.spc);
        seq.elementAt(0).accept(this);
        out.println();
        seq.elementAt(1).accept(this);
        out.print(spc.spc);
        seq.elementAt(2).accept(this);
        out.println();
      }
      else if((n.f0.which == 3) || (n.f0.which == 4) || (n.f0.which == 6)){
        out.print(spc.spc);
        n.f0.accept(this);
        out.println();
      }
      else{        
        n.f0.accept(this);
      }
      spc.updateSpc(-1);
   }

   //f0 -> "." "this"
   //     | "." AllocationExpression()
   //     | "[" "[" DistArraySecRefDimList() "]" "]"
   //     | "[" TripletOrExpressionList() "]"
   //     | "." < IDENTIFIER >
   //     | Arguments()
   public void visit(PrimarySuffix n) {
      out.print(spc.spc);
      out.print("PrimarySuffix");
      out.println();
      spc.updateSpc(+1);

      switch(n.f0.which){
        case 0:
          out.print(spc.spc);
          n.f0.accept(this);
          out.println();
          break;
        case 1:
          out.print(spc.spc);
          NodeSequence seq1 = (NodeSequence)n.f0.choice;
          seq1.elementAt(0).accept(this);
          out.println();
          seq1.elementAt(1).accept(this);
          break;
        case 2:
          out.print(spc.spc);
          NodeSequence seq2 = (NodeSequence)n.f0.choice;
          seq2.elementAt(0).accept(this);
          seq2.elementAt(1).accept(this);
          out.println();
          seq2.elementAt(2).accept(this);
          out.print(spc.spc);
          seq2.elementAt(3).accept(this);
          seq2.elementAt(4).accept(this);
          out.println();
          break;
        case 3:
          out.print(spc.spc);
          NodeSequence seq3 = (NodeSequence)n.f0.choice;
          seq3.elementAt(0).accept(this);
          out.println();
          seq3.elementAt(1).accept(this);
          out.print(spc.spc);
          seq3.elementAt(2).accept(this);
          break;
        case 4:
          out.print(spc.spc);
          n.f0.accept(this);
          out.println();
          break;
        case 5:
          n.f0.accept(this);
          break;
      default:
        throw new Error("Unexpected case.");
      }

      /*
      if(n.f0.which == 2){
        out.print(spc.spc);
        NodeSequence seq = (NodeSequence)n.f0.choice;
        seq.elementAt(0).accept(this);
        out.println();
        seq.elementAt(1).accept(this);
      }
      else if(n.f0.which == 3){
        out.print(spc.spc);
        NodeSequence seq = (NodeSequence)n.f0.choice;
        seq.elementAt(0).accept(this);
        out.println();
        seq.elementAt(1).accept(this);
        out.print(spc.spc);
        seq.elementAt(2).accept(this);
        out.println();
      }
      else if(n.f0.which == 5){
        n.f0.accept(this);        
      }
      else{        
        out.print(spc.spc);
        n.f0.accept(this);
        out.println();
      }
      */
      spc.updateSpc(-1);
   }

   //
   // f0 -> < INTEGER_LITERAL >
   //       | < FLOATING_POINT_LITERAL >
   //       | < CHARACTER_LITERAL >
   //       | < STRING_LITERAL >
   //       | BooleanLiteral()
   //       | NullLiteral()
   //
   public void visit(Literal n) {
      out.print(spc.spc);
      out.print("Literal");
      out.println(); 
      spc.updateSpc(+1);
      out.print(spc.spc);
      if(n.f0.which != 4 && n.f0.which != 5){
        n.f0.accept(this);
        out.println();
      }
      else n.f0.accept(this);
      spc.updateSpc(-1);
   }

   //
   // f0 -> "true"
   //       | "false"
   //
   public void visit(BooleanLiteral n) {
      out.print("BooleanLiteral");
      out.println();
      spc.updateSpc(+1);
      out.print(spc.spc);
      n.f0.accept(this);
      out.println();
      spc.updateSpc(-1);
   }

   //
   // f0 -> "null"
   //
   public void visit(NullLiteral n) {
      out.print("NullLiteral");
      out.println();
      spc.updateSpc(+1);
      out.print(spc.spc);
      n.f0.accept(this);
      out.println();
      spc.updateSpc(-1);
   }

   //
   // f0 -> "("
   // f1 -> [ ArgumentList() ]
   // f2 -> ")"
   //
   public void visit(Arguments n) {
      out.print(spc.spc);
      out.print("Arguments");
      out.println();
      spc.updateSpc(+1);
      n.f1.accept(this);
      spc.updateSpc(-1);
   }

   //
   // f0 -> Expression()
   // f1 -> ( "," Expression() )*
   //
   public void visit(ArgumentList n) {
      out.print(spc.spc);
      out.print("ArgumentList");
      out.println();
      spc.updateSpc(+1);
      n.f0.accept(this);

      if(n.f1.present()){
        for(int i=0; i<n.f1.size(); i++){
          NodeSequence seq = (NodeSequence)n.f1.elementAt(i);
          if(seq.elementAt(1) != null){
            seq.elementAt(1).accept(this);
            //out.println();
          }       
        } 
      }

      spc.updateSpc(-1);
   }

   //
   // f0 -> "new" PrimitiveType() ArrayShapes()
   //       | "new" Name() ( ArrayShapes() | Arguments() [ ClassBody() ] )
   //
   public void visit(AllocationExpression n) {
      out.print(spc.spc);
      out.print("AllocationExpression");
      out.println();
      spc.updateSpc(+1);
      out.print(spc.spc);
      NodeSequence seq = (NodeSequence)n.f0.choice;
      if ( n.f0.which == 0 ) {
         seq.elementAt(0).accept(this);
         out.println();
         seq.elementAt(1).accept(this);
         seq.elementAt(2).accept(this);
      }
      else {   // needs a little tweaking
         seq.elementAt(0).accept(this);
         out.println();
         seq.elementAt(1).accept(this);
         out.println();
         seq.elementAt(2).accept(this);
      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> ( AllocatedShape() )+ ( ArraySignature() )* [ "on" GroupExpression() ]
   //       | ( "[" "]" )+ ArrayInitializer()
   //       | "#" [ "on" GroupExpression() ]
   //
   public void visit(ArrayShapes n) {
      out.print(spc.spc);
      out.print("ArrayShapes");
      out.println();
      spc.updateSpc(+1);
      //out.print(spc.spc);
      //n.f0.accept(this);
      NodeSequence seq = (NodeSequence)n.f0.choice;
      if ( n.f0.which == 0 ) {
         seq.elementAt(0).accept(this);
         seq.elementAt(1).accept(this);

         if(((NodeOptional)seq.elementAt(2)).present()){
           NodeOptional no = (NodeOptional)seq.elementAt(2);
           NodeSequence seq2 = (NodeSequence)no.node;
           out.print(spc.spc);
           seq2.elementAt(0).accept(this);
           out.println();
           seq2.elementAt(1).accept(this);
         }
      }
      else if ( n.f0.which == 1 ) {
         seq.elementAt(0).accept(this);
         seq.elementAt(1).accept(this);
      }

      else {   // needs a little tweaking
         seq.elementAt(0).accept(this);
         out.println();
         seq.elementAt(1).accept(this);
      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> "[" "[" Expression() ( "," Expression() )* "]" "]"
   //       | "[" Expression() "]"
   //
   public void visit(AllocatedShape n) {
      out.print(spc.spc);
      out.print("AllocatedShape");
      out.println();
      spc.updateSpc(+1);
      //out.print(spc.spc);
      //n.f0.accept(this);
      NodeSequence seq = (NodeSequence)n.f0.choice;
      if ( n.f0.which == 0 ) {
         out.print(spc.spc);
         seq.elementAt(0).accept(this);
         seq.elementAt(1).accept(this);
         out.println();
         seq.elementAt(2).accept(this);
         seq.elementAt(3).accept(this);
         out.print(spc.spc);
         seq.elementAt(4).accept(this);
         seq.elementAt(5).accept(this);
         out.println();
      }
      else {   // needs a little tweaking
         out.print(spc.spc);
         seq.elementAt(0).accept(this);
         out.println();
         //out.println("[");
         seq.elementAt(1).accept(this);
         out.print(spc.spc);
         seq.elementAt(2).accept(this);
         out.println();
      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> PrimaryExpression()
   // f1 -> ( "/" PrimaryExpression() )*
   //
   public void visit(GroupExpression n) {
      out.print(spc.spc);
      out.print("GroupExpression");
      out.println();
      spc.updateSpc(+1);
      n.f0.accept(this);
      if(n.f1.present()){
        for(int i=0; i<n.f1.size(); i++){
          NodeSequence seq = (NodeSequence)n.f1.elementAt(i);
          out.print(spc.spc);
          seq.elementAt(0).accept(this);
          out.println();
          seq.elementAt(1).accept(this);
        } 

      }
      spc.updateSpc(-1);

   }

   //
   // f0 -> LabeledStatement()
   //       | Block()
   //       | EmptyStatement()
   //       | StatementExpression() ";"
   //       | SwitchStatement()
   //       | IfStatement()
   //       | WhileStatement()
   //       | DoStatement()
   //       | ForStatement()
   //       | BreakStatement()
   //       | ContinueStatement()
   //       | ReturnStatement()
   //       | ThrowStatement()
   //       | SynchronizedStatement()
   //       | TryStatement()
   //       | OnStatement()
   //       | AtStatement()
   //       | OverallStatement()
   //
   public void visit(Statement n) {
      out.print(spc.spc);
      out.print("Statement");
      out.println();
      spc.updateSpc(+1);
      if(n.f0.which == 3){
        NodeSequence seq = (NodeSequence)n.f0.choice;
        seq.elementAt(0).accept(this);
      }
      else{
        n.f0.accept(this);
      }   
      spc.updateSpc(-1);
   }

   //
   // f0 -> < IDENTIFIER >
   // f1 -> ":"
   // f2 -> Statement()
   //
   public void visit(LabeledStatement n) {
      out.print(spc.spc);
      out.print("LabeledStatement");
      out.println();
      spc.updateSpc(+1);
      out.print(spc.spc);
      n.f0.accept(this);
      out.println();
      out.print(spc.spc);
      n.f1.accept(this);
      out.println();
      n.f2.accept(this);
      spc.updateSpc(-1);
   }

   //
   // f0 -> "{"
   // f1 -> ( BlockStatement() )*
   // f2 -> "}"
   //
   public void visit(Block n) {
      out.print(spc.spc);
      out.print("Block");
      out.println();
      spc.updateSpc(+1);
      n.f1.accept(this);
      spc.updateSpc(-1);
   }

   //
   // f0 -> LocalVariableDeclaration() ";"
   //       | Statement()
   //       | UnmodifiedClassDeclaration()
   //
   public void visit(BlockStatement n) {
      out.print(spc.spc);
      out.print("BlockStatement");
      out.println();
      spc.updateSpc(+1);
      if(n.f0.which == 0) {
         NodeSequence seq = (NodeSequence)n.f0.choice;
         seq.elementAt(0).accept(this);
      }
      else{
         n.f0.accept(this);
      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> [ "final" ]
   // f1 -> Type()
   // f2 -> VariableDeclarator()
   // f3 -> ( "," VariableDeclarator() )*
   //
   public void visit(LocalVariableDeclaration n) {
      out.print(spc.spc);
      out.print("LocalVariableDeclaration");
      out.println();
      spc.updateSpc(+1);
      if(n.f0.present()){
        out.print(spc.spc);
        n.f0.accept(this);
        out.println();
      }
      n.f1.accept(this);
      n.f2.accept(this);
      if(n.f3.present()){
        for(int i=0; i<n.f3.size(); i++){
          NodeSequence seq = (NodeSequence)n.f3.elementAt(i);
          seq.elementAt(1).accept(this);
        } 

      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> ";"
   //
   public void visit(EmptyStatement n) {
      out.print(spc.spc);
      out.print("EmptyStatement");
      out.println();
   }

   //
   // f0 -> PreIncrementExpression()
   //       | PreDecrementExpression()
   //       | PrimaryExpression() [ "++" | "--" | AssignmentOperator() Expression() ]
   //

   public void visit(StatementExpression n) {
      out.print(spc.spc);
      out.print("StatementExpression");
      out.println();
      spc.updateSpc(+1);

      if(n.f0.which == 2){
        NodeSequence seq = (NodeSequence)n.f0.choice;
        seq.elementAt(0).accept(this);
        if(((NodeOptional)seq.elementAt(1)).present()){
          NodeOptional no = (NodeOptional)seq.elementAt(1);
          NodeChoice nc = (NodeChoice)no.node;
          if(nc.which == 0 || nc.which == 1){
            out.print(spc.spc);
            nc.accept(this);
            out.println();
          }        
          else{
            nc.accept(this);
          }
        }
      }
      else{
        n.f0.accept(this);     
      }   

     spc.updateSpc(-1);
   }

   //
   // f0 -> "switch"
   // f1 -> "("
   // f2 -> Expression()
   // f3 -> ")"
   // f4 -> "{"
   // f5 -> ( SwitchLabel() ( BlockStatement() )* )*
   // f6 -> "}"
   //
   public void visit(SwitchStatement n) {
      out.print(spc.spc);
      out.print("SwitchStatement");
      out.println();
      spc.updateSpc(+1);
      out.print(spc.spc);
      n.f0.accept(this);
      out.println();
      n.f2.accept(this);

      if(n.f5.present()){
        for(int i=0; i<n.f5.size(); i++){
          NodeSequence seq = (NodeSequence)n.f5.elementAt(i);
          seq.elementAt(0).accept(this);
          if(((NodeListOptional)seq.elementAt(1)).present()){
            for(int j=0; j<((NodeListOptional)seq.elementAt(1)).size(); j++){
              ((NodeListOptional)seq.elementAt(1)).elementAt(j).accept(this);
            }
          }       
        } 
        
      }

      spc.updateSpc(-1);
   }

   //
   // f0 -> "case" Expression() ":"
   //       | "default" ":"
   //
   public void visit(SwitchLabel n) {
      out.print(spc.spc);
      out.print("SwitchLabel");
      out.println();
      spc.updateSpc(+1);
  
      if(n.f0.which == 0){
        NodeSequence seq = (NodeSequence)n.f0.choice;
        out.print(spc.spc);
        seq.elementAt(0).accept(this);
        out.println();
        seq.elementAt(1).accept(this);
      }    
      else{
        NodeSequence seq = (NodeSequence)n.f0.choice;
        out.print(spc.spc);
        seq.elementAt(0).accept(this);
        out.println();
      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> "if"
   // f1 -> "("
   // f2 -> Expression()
   // f3 -> ")"
   // f4 -> Statement()
   // f5 -> [ "else" Statement() ]
   //
   public void visit(IfStatement n) {
      out.print(spc.spc);
      out.print("IfStatement");
      out.println();
      spc.updateSpc(+1);
      out.print(spc.spc);
      n.f0.accept(this);
      out.println();
      n.f2.accept(this);

      if ( n.f4.f0.which != 1 )     // only indent if it's not a Block
         spc.updateSpc(+1);

      n.f4.accept(this);

      if ( n.f4.f0.which != 1 )
         spc.updateSpc(-1);

      if ( n.f5.present() ) {

         if (((Statement)((NodeSequence)n.f5.node).elementAt(1)).f0.which != 1)
            spc.updateSpc(+1);

//         n.f5.accept(this);

         NodeSequence seq = (NodeSequence)n.f5.node;
         out.print(spc.spc);
         seq.elementAt(0).accept(this);
         out.println();
         seq.elementAt(1).accept(this);

         if (((Statement)((NodeSequence)n.f5.node).elementAt(1)).f0.which != 1)
            spc.updateSpc(-1);

      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> "while"
   // f1 -> "("
   // f2 -> Expression()
   // f3 -> ")"
   // f4 -> Statement()
   //
   public void visit(WhileStatement n) {
      out.print(spc.spc);
      out.print("WhileStatement");
      out.println();
      spc.updateSpc(+1);
      out.print(spc.spc);
      n.f0.accept(this);
      out.println();
      n.f2.accept(this);

      if ( n.f4.f0.which != 1 )
         spc.updateSpc(+1);

      n.f4.accept(this);

      if ( n.f4.f0.which != 1 )
         spc.updateSpc(-1);
      spc.updateSpc(-1);
   }

   //
   // f0 -> "do"
   // f1 -> Statement()
   // f2 -> "while"
   // f3 -> "("
   // f4 -> Expression()
   // f5 -> ")"
   // f6 -> ";"
   //
   public void visit(DoStatement n) {
      out.print(spc.spc);
      out.print("DoStatement");
      out.println();
      spc.updateSpc(+1);
      out.print(spc.spc);
      n.f0.accept(this);
      out.println();

      if ( n.f1.f0.which != 1 )
         spc.updateSpc(+1);

      n.f1.accept(this);

      if ( n.f1.f0.which != 1 )
         spc.updateSpc(-1);

      out.print(spc.spc);
      n.f2.accept(this);
      out.println();
      n.f4.accept(this);
      spc.updateSpc(-1);
   }

   //
   // f0 -> "for"
   // f1 -> "("
   // f2 -> [ ForInit() ]
   // f3 -> ";"
   // f4 -> [ Expression() ]
   // f5 -> ";"
   // f6 -> [ ForUpdate() ]
   // f7 -> ")"
   // f8 -> Statement()
   //
   public void visit(ForStatement n) {
      out.print(spc.spc);
      out.print("ForStatement");
      out.println();
      spc.updateSpc(+1);
      out.print(spc.spc);
      n.f0.accept(this);
      out.println();
      if(n.f2.present()) n.f2.accept(this);
      if(n.f4.present()) n.f4.accept(this);
      if(n.f6.present()) n.f6.accept(this);

      if ( n.f8.f0.which != 1 )
         spc.updateSpc(+1);

      n.f8.accept(this);

      if ( n.f8.f0.which != 1 )
         spc.updateSpc(-1);

      spc.updateSpc(-1);
   }

   //
   // f0 -> LocalVariableDeclaration()
   //       | StatementExpressionList()
   //
   public void visit(ForInit n) {
      out.print(spc.spc);
      out.print("ForInit");
      out.println();
      spc.updateSpc(+1);
      n.f0.accept(this);
      spc.updateSpc(-1);
   }

   //
   // f0 -> StatementExpression()
   // f1 -> ( "," StatementExpression() )*
   //
   public void visit(StatementExpressionList n) {
      out.print(spc.spc);
      out.print("StatementExpressionList");
      out.println();
      spc.updateSpc(+1);
      n.f0.accept(this);
      if(n.f1.present()){
        for(int i=0; i<n.f1.size(); i++){
          NodeSequence seq = (NodeSequence)n.f1.elementAt(i);
          seq.elementAt(1).accept(this);
        } 
      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> StatementExpressionList()
   //
   public void visit(ForUpdate n) {
      out.print(spc.spc);
      out.print("ForUpdate");
      out.println();
      spc.updateSpc(+1);
      n.f0.accept(this);
      spc.updateSpc(-1);
   }

   //
   // f0 -> "break"
   // f1 -> [ < IDENTIFIER >  ]
   // f2 -> ";"
   //
   public void visit(BreakStatement n) {
      out.print(spc.spc);
      out.print("BreakStatement");
      out.println();
      spc.updateSpc(+1);
      out.print(spc.spc);
      n.f0.accept(this);
      out.println();
      if(n.f1.present()){
        out.print(spc.spc);
        n.f1.accept(this);
        out.println();
      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> "continue"
   // f1 -> [ < IDENTIFIER >  ]
   // f2 -> ";"
   //
   public void visit(ContinueStatement n) {
      out.print(spc.spc);
      out.print("ContinueStatement");
      out.println();
      spc.updateSpc(+1);
      out.print(spc.spc);
      n.f0.accept(this);
      out.println();
      if(n.f1.present()){
        out.print(spc.spc);
        n.f1.accept(this);
        out.println();
      }

      spc.updateSpc(-1);
   }

   //
   // f0 -> "return"
   // f1 -> [ Expression() ]
   // f2 -> ";"
   //
   public void visit(ReturnStatement n) {
      out.print(spc.spc);
      out.print("ReturnStatement");
      out.println();
      spc.updateSpc(+1);
      out.print(spc.spc);
      n.f0.accept(this);
      out.println();
      if(n.f1.present()) n.f1.accept(this);
      spc.updateSpc(-1);
   }

   //
   // f0 -> "throw"
   // f1 -> Expression()
   // f2 -> ";"
   //
   public void visit(ThrowStatement n) {
      out.print(spc.spc);
      out.print("ThrowStatement");
      out.println();
      spc.updateSpc(+1);
      out.print(spc.spc);
      n.f0.accept(this);
      out.println();
      n.f1.accept(this);
      spc.updateSpc(-1);
   }

   //
   // f0 -> "synchronized"
   // f1 -> "("
   // f2 -> Expression()
   // f3 -> ")"
   // f4 -> Block()
   //
   public void visit(SynchronizedStatement n) {
      out.print(spc.spc);
      out.print("SynchronizedStatement");
      out.println();
      spc.updateSpc(+1);
      out.print(spc.spc);
      n.f0.accept(this);
      out.println();
      n.f2.accept(this);
      n.f4.accept(this);
      spc.updateSpc(-1);
   }

   //
   // f0 -> "try"
   // f1 -> Block()
   // f2 -> ( "catch" "(" FormalParameter() ")" Block() )*
   // f3 -> [ "finally" Block() ]
   //
   public void visit(TryStatement n) {
      out.print(spc.spc);
      out.print("TryStatement");
      out.println();
      spc.updateSpc(+1);
      out.print(spc.spc);
      n.f0.accept(this);
      out.println();
      n.f1.accept(this);

      if(n.f2.present()){
        for(int i=0; i<n.f2.size(); i++){
          NodeSequence seq = (NodeSequence)n.f2.elementAt(i);
          out.print(spc.spc);
          out.print("CatchStatement");
          out.println();
          spc.updateSpc(+1);

          out.print(spc.spc);
          seq.elementAt(0).accept(this);
          out.println();
          seq.elementAt(2).accept(this);
          seq.elementAt(4).accept(this);

          spc.updateSpc(-1);
        } 
        
      }
      if( n.f3.present() ) {
        NodeSequence seq = (NodeSequence)n.f3.node;
        out.print(spc.spc);
        seq.elementAt(0).accept(this);
        out.println();
        seq.elementAt(1).accept(this);
      }
      spc.updateSpc(-1);
   }

   //
   // f0 -> "on"
   // f1 -> "("
   // f2 -> Expression()
   // f3 -> ")"
   // f4 -> Statement()
   //
   public void visit(OnStatement n) {
      out.print(spc.spc);
      out.print("OnStatement");
      out.println();
      spc.updateSpc(+1);
      out.print(spc.spc);
      n.f0.accept(this);
      out.println();
      n.f2.accept(this);

      if ( n.f4.f0.which != 1 )
         spc.updateSpc(+1);

      n.f4.accept(this);

      if ( n.f4.f0.which != 1 )
         spc.updateSpc(-1);
      spc.updateSpc(-1);
   }

   //
   // f0 -> "at"
   // f1 -> "("
   // f2 -> NamedLocationList()
   // f3 -> ")"
   // f4 -> Statement()
   //
   public void visit(AtStatement n) {
      out.print(spc.spc);  
      out.print("AtStatement");  
      out.println();  
       
      spc.updateSpc(+1);  
 
      out.print(spc.spc); 
      n.f0.accept(this); 
      out.println(); 
 
      n.f2.accept(this); 
 
      //out.print(spc.spc); 
      //n.f3.accept(this); 
      //out.println(); 
 
      n.f4.accept(this); 
 
      spc.updateSpc(-1); 

   }

  /*
   //
   // f0 -> "at"
   // f1 -> "("
   // f2 -> <IDENTIFIER>
   // f3 -> "="
   // f4 -> Expression()
   // f5 -> ")"
   // f6 -> Statement()
   //
   public void visit(AtStatement n) {
      out.print(spc.spc); 
      out.print("AtStatement");
      out.println(); 
      spc.updateSpc(+1);
      out.print(spc.spc);
      n.f0.accept(this);
      out.println();
      out.print(spc.spc);
      n.f2.accept(this);
      out.println();
      out.print(spc.spc);
      n.f3.accept(this);
      out.println();
      n.f4.accept(this);

      if ( n.f6.f0.which != 1 )
         spc.updateSpc(+1);

      n.f6.accept(this);

      if ( n.f6.f0.which != 1 )
         spc.updateSpc(-1);
      spc.updateSpc(-1);
   }
  */


   //
   // f0 -> NamedLocation()
   // f1 -> ( "," NamedLocation() )*
   //
   public void visit(NamedLocationList n) {
      out.print(spc.spc); 
      out.print("NamedLocationList"); 
      out.println(); 
      spc.updateSpc(+1); 
      n.f0.accept(this); 
 
      if(n.f1.present()){ 
        for(int i=0; i<n.f1.size(); i++){ 
          NodeSequence seq = (NodeSequence)n.f1.elementAt(i); 
          if(seq.elementAt(1) != null){ 
            seq.elementAt(1).accept(this); 
            //out.println(); 
          }        
        }  
      } 
 
      spc.updateSpc(-1); 
   }

   /**
    * <PRE>
    * f0 -> &lt;IDENTIFIER&gt;
    * f1 -> "="
    * f2 -> PrimaryExpression()
    * </PRE>
    */
   public void visit(NamedLocation n) {
      out.print(spc.spc);
      out.print("NamedLocation");
      out.println();

      spc.updateSpc(+1);

      out.print(spc.spc);
      n.f0.accept(this);
      out.println();

      out.print(spc.spc);
      n.f1.accept(this);
      out.println();

      n.f2.accept(this);

      spc.updateSpc(-1);
   }


   //
   // f0 -> "overall"
   // f1 -> "("
   // f2 -> NamedLocationRangeList()
   // f3 -> ")"
   // f4 -> Statement()
   //
   public void visit(OverallStatement n) {
      out.print(spc.spc); 
      out.print("OverallStatement"); 
      out.println(); 
      
      spc.updateSpc(+1); 

      out.print(spc.spc);
      n.f0.accept(this);
      out.println();

      n.f2.accept(this);

      //out.print(spc.spc);
      //n.f3.accept(this);
      //out.println();

      n.f4.accept(this);

      spc.updateSpc(-1);
   }

   //
   // f0 -> NamedLocationRange()
   // f1 -> ( "," NamedLocationRange() )*
   //
   public void visit(NamedLocationRangeList n) {
     //n.f0.accept(this);
     //n.f1.accept(this);
      out.print(spc.spc); 
      out.print("NamedLocationRangeList"); 
      out.println(); 
      spc.updateSpc(+1); 
      n.f0.accept(this); 
 
      if(n.f1.present()){ 
        for(int i=0; i<n.f1.size(); i++){ 
          NodeSequence seq = (NodeSequence)n.f1.elementAt(i); 
          if(seq.elementAt(1) != null){ 
            seq.elementAt(1).accept(this); 
            //out.println(); 
          }        
        }  
      } 
 
      spc.updateSpc(-1); 
   }

   //
   // f0 -> <IDENTIFIER>
   // f1 -> "="
   // f2 -> PrimaryExpression()
   // f3 -> "for"
   // f4 -> Triplet()
   //
   public void visit(NamedLocationRange n) {
      //n.f0.accept(this);
      //n.f1.accept(this);
      //n.f2.accept(this);
      //n.f3.accept(this);
      //n.f4.accept(this);
      out.print(spc.spc);
      out.print("NamedLocationRange");
      out.println();

      spc.updateSpc(+1);

      out.print(spc.spc);
      n.f0.accept(this);
      out.println();

      out.print(spc.spc);
      n.f1.accept(this);
      out.println();

      n.f2.accept(this);
      
      out.print(spc.spc);
      n.f3.accept(this);
      out.println();
      
      n.f4.accept(this);

      spc.updateSpc(-1);
   }

}


